unit Liikkuva;
{
  Olio, joka on tehty kuljettamaan muita olioita paikasta toiseen.
  Vesa Lappalainen 7.11.1997

  Liikuttaa olion olion nykyiset paikasta joko annetun WinControllin
  paikkaan pt (KuljetaPt) tai annettun kontrollin plle (Kuljeta).
  Jos kytetn funktiota tai metodia KuljetaPT, pit sek lhtpaikka
  ett saapumispaikka olla saman kontrollin alueella.
  Saadaan tapahtuma jokaisesta liikahduksesta (OnMove) sek perille
  saapumisesta (OnArrive).
  Matka tehdn joko ajassa Time tai nopeudella SpeedPts (pxl/s)
  (kytetn sit joka <> 0). Matkalla otetaan Steps askelta
  tai siirtojen vli on Interval (kytetn sit, joka <>0).

  Kytt helpottamaan on tehty kaksi funktiota, jotka luovat liikuvan
  olio, siirtvt siirrettvn olien liikkuvan sislle ja ilmoittavat
  lopuksi kun siirto on valmis.

  function KuljetaPt(OWhat : TControl; f:TWinControl;ToPt:TPoint;
                    arrive:TLiikkuvaMeth;data : integer; v:integer) : boolean;
  function Kuljeta(OWhat,OTo : TControl; arrive:TLiikkuvaMeth;
                    data:integer; v:integer) : boolean;
  Parametrit:
    OWhat    - siirrettv olio
    f        - ikkuna jonka sisll siirto tehdn
    ToPt     - koordinaatti, johon siirto loppuu
    arrive   - metodi, joka ksittelee valmiin siirron
    data     - vapaasti kytettv sana, joka saadaan kyttn siirron loputtua
    v        - jos >0 siirron nopeus pxl/s
               jos <0 siirron kesto ms (abs(v))
    OTo      - olio johon siirto pttyy           

  Muutoksia 11.11.1997/vl
   + olion Parent := nil ennen Arrive-metodin kutsumista
}

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  ExtCtrls;

type
  TLiikkuvaMeth = procedure (o:TControl;data:integer) of object;
  TLiikkuvaMoveMeth = procedure (o:TControl;pt:TPoint;data:integer) of object;

  TLiikkuva = class(TWinControl)
  private
    FTimer : TTimer;
    FOnArrive : TLiikkuvaMeth;
    FOnMove   : TLiikkuvaMoveMeth;
    FObject   : TControl;
    FDpt      : TPoint;
    FTo       : TPoint;
    FFrom     : TPoint;
    FData     : Integer;
    FOrigo    : TPoint;

    FStep     : integer;
    FSteps    : integer;
    FTime     : integer;
    FSpeedPts : integer;
    FInterval : integer;

    FUseSteps : integer;

    FAutoDestroy : boolean;
    procedure FTimerTimer(Sender: TObject);
  protected
  public
    constructor Create(AOwner:TComponent);                  override;
    destructor  Destroy;                                    override;
    function KuljetaPt(OWhat : TControl; f:TWinControl;ToPt:TPoint;
                       data : integer) : boolean; virtual;
    function Kuljeta(OWhat,OTo : TControl; data:integer):boolean; virtual;
    function Moving : boolean;                              virtual;
  published
    property OnArrive : TLiikkuvaMeth read FOnArrive write FOnArrive;
    property OnMove   : TLiikkuvaMoveMeth read FOnMove write FOnMove;
    property Steps : integer read FSteps write FSteps default 20;
    property TimeMs : integer read FTime write FTime default 1000;
    property SpeedPts : integer read FSpeedPts write FSpeedPts default 0;
    property Interval : integer read FInterval write FInterval default 0;
  end;

function KuljetaPt(OWhat : TControl; f:TWinControl;ToPt:TPoint;
                   arrive:TLiikkuvaMeth;data : integer; v:integer) : boolean;
function Kuljeta(OWhat,OTo : TControl; arrive:TLiikkuvaMeth;
                 data:integer; v:integer) : boolean;

procedure Register;

implementation

constructor TLiikkuva.Create(AOwner:TComponent);
begin
  inherited;
//  BevelInner := bvNone;
//  BevelOuter := bvNone;
//  ParentColor := true;
  FTimer := TTimer.Create(self);
  FTimer.Enabled := false;
  FTimer.OnTimer := FTimerTimer;
  FAutoDestroy := false;
  FOnMove := nil;
  FOnArrive := nil;
  FOrigo.x := 0; FOrigo.y := 0;
  FInterval := 0;
  FSteps    := 20;
  FSpeedPts := 0;
  FTime     := 1000;
end;

destructor  TLiikkuva.Destroy;
begin
  FTimer.Free;
  inherited;
end;

function TLiikkuva.Moving : boolean;
begin
  Result := FTimer.Enabled;
end;

function TLiikkuva.KuljetaPt(OWhat : TControl; f:TWinControl;ToPt:TPoint;
                             data : integer) : boolean;
var dist,t,dt : integer;
begin
  Result := false;
  if ( Moving ) then exit;
  FObject := OWhat;
  FFrom := f.ScreenToClient(FObject.ClientToScreen(FOrigo));
//  FObject.Owner := self;
  FObject.Parent := self;
  FObject.Top := 0;
  FObject.Left := 0;
  FData := data;
  Width := FObject.Width;
  Height := FObject.Height;
  Parent := f;

  FTo := ToPt;
  Left := FFrom.x;
  Top := FFrom.y;

  FUseSteps := 0;  if ( Steps <> 0 ) then FUseSteps := Steps;

  FDpt.x := (FTo.x - FFrom.x);
  FDpt.y := (FTo.y - FFrom.y);
  Dist := round(sqrt(1.0*FDpt.x*FDpt.x + 1.0*FDpt.y*FDpt.y));

  t := TimeMs;
  if ( SpeedPts <> 0 ) then  // v = s/t
    t := Round(1000*(dist/SpeedPts));
  if ( t = 0 ) then t := 1000;

  dt := 100;
  if ( Interval > 0 ) then dt := Interval;
  if ( FUseSteps <> 0 ) then dt := t div FUseSteps
  else FUseSteps := t div dt;

  if ( dt < 30 ) then begin
    dt := 30;
    FUseSteps := t div dt;
  end;  

  If FUseSteps = 0 then FUseSteps := 1;
  FStep := 0;
  FTimer.Interval := dt;
  FTimer.Enabled := true;
  Visible := true;
end;

function TLiikkuva.Kuljeta(OWhat, OTo : TControl; data : integer) : boolean;
var p : TObject; f:TForm; pt:TPoint;
begin
  Result := false;
  if ( Moving ) then exit;
  p := OTo;
  while ( p<> nil ) and (not ( p is TForm )) do
    p := (p as TControl).Parent;
  if ( p = nil ) then exit;
  f := p as TForm;

  pt := f.ScreenToClient(OTo.ClientToScreen(FOrigo));

  Result := KuljetaPt(OWhat,f,pt,data)
end;

procedure TLiikkuva.FTimerTimer(Sender: TObject);
var pt:TPoint;
begin
  Left := FFrom.x + Round(FStep*(FDpt.x/FUseSteps));
  Top  := FFrom.y + Round(FStep*(FDpt.y/FUseSteps));
  pt.x := Left;
  pt.y := Top;
  inc(FStep);
  if ( Assigned(FOnMove) ) then OnMove(FObject,pt,FData);
  if ( FStep <= FUseSteps ) then exit;
  FTimer.Enabled := false;
  FObject.Parent := nil;
  if ( Assigned(FOnArrive) ) then OnArrive(FObject,FData);
  if ( FAutoDestroy ) then Free;
end;

function InitLiikkuva(var l:TLiikkuva;f:TComponent;
                      arrive:TLiikkuvaMeth; v:integer) : boolean;
begin
  Result := false;
  l := TLiikkuva.Create(f);
  if ( l = nil ) then exit;
  l.Visible := false;
  l.FAutoDestroy := true;
  l.OnArrive := arrive;
  if ( v > 0 ) then l.SpeedPts := v
               else l.TimeMs := abs(v);
  Result := true;
end;


function KuljetaPt(OWhat : TControl; f:TWinControl;ToPt:TPoint;
                   arrive:TLiikkuvaMeth;data : integer; v:integer) : boolean;
var liikkuva : TLiikkuva;
begin
  Result := false;
  if not InitLiikkuva(Liikkuva,f,arrive,v) then exit;
  Result := liikkuva.KuljetaPt(OWhat,f,ToPt,data);
end;

function Kuljeta(OWhat,OTo : TControl; arrive:TLiikkuvaMeth; data:integer; v:integer) : boolean;
var liikkuva : TLiikkuva;
begin
  Result := false;
  if not InitLiikkuva(Liikkuva,OTo.Owner,arrive,v) then exit;
  Result := liikkuva.Kuljeta(OWhat,OTo,data);
end;

procedure Register;
begin
  RegisterComponents('GKO97', [TLiikkuva]);
end;

end.
